#ifndef __GST_OBJECT_SET_PROPERTIES_FROM_URI_QUERY_H__
#define __GST_OBJECT_SET_PROPERTIES_FROM_URI_QUERY_H__

#include <gst/gst.h>

void
gst_object_set_properties_from_uri_query (GObject * obj, const GstUri * uri);

#endif
